@@ -1,8 +1,8 @@
 """ _selector_events.py - AsyncSlot based on SelectorEventLoop """
 
 import asyncio
-import selectors
 import concurrent.futures
+import selectors
 import threading
 import weakref
 from typing import List, Optional, Tuple
@@ -13,6 +13,7 @@ __all__ = 'AsyncSlotSelectorEventLoop', 'AsyncSlotSelectorEventLoopPolicy',
 
 
 class AsyncSlotSelector(selectors.BaseSelector):
+
     def __init__(self, selector: selectors.BaseSelector,
                  write_to_self: weakref.WeakMethod):
         super().__init__()
@@ -20,21 +21,24 @@ class AsyncSlotSelector(selectors.BaseSelector):
         self._write_to_self = write_to_self
         self._executor = concurrent.futures.ThreadPoolExecutor(max_workers=1)
         self._select_future: Optional[concurrent.futures.Future] = None
-        self._select_done = threading.Event()
+        self._idle = threading.Event()
+        self._idle.set()
         self._notifier: Optional[AsyncSlotNotifier] = None
+        self._closed = False
 
     def set_notifier(self, notifier: Optional[AsyncSlotNotifier]) -> None:
         self._unblock_if_blocked()
         self._notifier = notifier
 
     def _unblock_if_blocked(self):
-        if self._select_future is not None and not self._select_done.is_set():
+        assert not self._closed, 'selector already closed'
+        if not self._idle.is_set():
             write_to_self = self._write_to_self()
             assert write_to_self is not None, (
                 'AsyncSlotEventLoop is supposed to close AsyncSlotSelector '
                 'before being deleted')
             write_to_self()
-            self._select_future.result()  # waits
+            self._idle.wait()
 
     def register(self, fileobj, events, data=None):
         self._unblock_if_blocked()
@@ -50,50 +54,69 @@ class AsyncSlotSelector(selectors.BaseSelector):
 
     def select(self, timeout: Optional[float] = None) \
             -> List[Tuple[selectors.SelectorKey, int]]:
+        assert not self._closed, 'selector already closed'
 
+        # If the last call to select() raised AsyncSlotYield, the caller
+        # (from _run_once) should only call us again after receiving a
+        # notification from us, and we only send the notification after
+        # entering IDLE state.
+        assert self._idle.is_set(), 'unexpected select'
+
+        # Return previous select() result (or exception) if there is one.
         if self._select_future is not None:
-            # A prior select() call was submitted to the executor.  We only
-            # submit a select() call if _run_once() calls us with a positive
-            # timeout, which can only happen if there are no ready tasks to
-            # execute.  That this method is called again means _run_once is
-            # run again, which can only happen if we asked it to by emitting
-            # the notified signal of __notifier.
-            assert self._select_done.is_set(), 'unexpected select'
             try:
                 return self._select_future.result()
             finally:
-                self._select_done.clear()
                 self._select_future = None
 
-        # Try select with zero timeout.  If any IO is ready or if the caller
-        # does not require IO to be ready, return that.
+        # Perform normal (blocking) select if no notifier is set.
+        if self._notifier is None:
+            return self._selector.select(timeout)
+
+        # Perform normal select if timeout is zero.
+        if timeout == 0:
+            return self._selector.select(timeout)
+
+        # Try select with zero timeout, and return if any IO is ready.
         event_list = self._selector.select(0)
-        if event_list or (timeout is not None and timeout <= 0):
+        if event_list:
             return event_list
 
         # No IO is ready and caller wants to wait.  select() in a separate
         # thread and tell the caller to yield.
-        assert self._notifier is not None, 'missing set_notifier'
-        self._select_future = self._executor.submit(self._select, timeout)
-        raise AsyncSlotYield
+        self._idle.clear()
+        try:
+            self._select_future = self._executor.submit(self._select, timeout)
+        except BaseException:
+            # Should submit() raise, we assume no task is spawned.
+            self._idle.set()
+            raise
+        else:
+            raise AsyncSlotYield
 
     def _select(self, timeout):
-        # Any exception raised by self._selector.select() is stored in the
-        # future and propagated to the next _run_once() call.
+        # Make a copy of self._notifier because it may be altered by
+        # set_notifier immediately after self._idle is set.
+        notifier = self._notifier
         try:
             return self._selector.select(timeout)
         finally:
-            self._select_done.set()
-            self._notifier.notify()
+            self._idle.set()
+            notifier.notify()
 
     def close(self) -> None:
-        self._unblock_if_blocked()
-        self._executor.shutdown()
-        self._selector.close()
-        # If close() is called before consuming the result of select(), the
-        # result is dropped.
-        self._select_future = None
-        self._notifier = None
+        # close() is called when the loop is being closed, and the loop
+        # can only be closed when it is in STOPPED state.  In this state
+        # the selector must be idle.  In addition, the self pipe is
+        # closed before closing the selector, so write_to_self cannot be
+        # used at this point.
+        if not self._closed:
+            assert self._idle.is_set(), 'unexpected close'
+            self._executor.shutdown()
+            self._selector.close()
+            self._select_future = None
+            self._notifier = None
+            self._closed = True
 
     def get_key(self, fileobj):
         self._unblock_if_blocked()
@@ -107,10 +130,10 @@ class AsyncSlotSelector(selectors.BaseSelector):
 class AsyncSlotSelectorEventLoop(AsyncSlotBaseEventLoop,
                                  asyncio.SelectorEventLoop):
 
-    def __init__(self):
+    def __init__(self, *, standalone=True):
         selector = AsyncSlotSelector(selectors.DefaultSelector(),
                                      weakref.WeakMethod(self._write_to_self))
-        super().__init__(selector)
+        super().__init__(selector, standalone=standalone)
 
 
 class AsyncSlotSelectorEventLoopPolicy(asyncio.events.BaseDefaultEventLoopPolicy):
